/*
 * Copyright (C) 2020 Christian Zietz
 *
 * This file is distributed under the GPL, version 2 or at your
 * option any later version.  See doc/license.txt for details.
 */

#include <osbind.h>
#include <mint/sysvars.h>

#include "rtc_driver.h"

/* Use the fact that gcc can have assembler code outside of functions. */
__asm (
    /* XBRA header. */
    ".ascii \"XBRA\"\n\t"
    ".ascii \"SRTC\"\n"
    "_xbios_old_trap:\n\t"
    ".dc.l 0\n"

    "_xbios_trap:\n\t"
    /* Get pointer to arguments. */
    "lea    6(sp),a0\n\t"
    "tst.w  0x59e.w\n\t" /* _longframe */
    "jbeq   is68000\n\t"
    "addq.l #2,a0\n"
    "is68000:\n\t"
    /* Were we called from user mode? Then arguments are on user stack */
    "btst   #(13-8),(sp)\n\t"
    "jbne   fromsuper\n\t"
    "move.l usp,a0\n"
    "fromsuper:\n\t"

    /* Call to Gettime? */
    "cmpi.w #23,(a0)\n\t"
    "jbeq   gettime\n\t"
    /* Call to Settime? */
    "cmpi.w #22,(a0)\n\t"
    "jbeq   settime\n\t"
    /* No: call old handler. */
    "move.l  _xbios_old_trap,-(sp)\n\t"
    "rts\n"

    /* Handle Gettime */
    "gettime:\n"
    "move.l #0,-(sp)\n\t"
    /* No need to save registers; A0-A2 and D0-D2 are allowed to be clobbered by XBIOS trap. */
    "jbsr   _rtc_gettime\n\t"
    "addq.l #4,sp\n\t"
    /* Result is already in D0 */
    "rte\n"

    /* Handle Settime */
    "settime:\n\t"
    "move.l 2(a0),-(sp)\n\t"
    /* No need to save registers; A0-A2 and D0-D2 are allowed to be clobbered by XBIOS trap. */
    "jbsr   _rtc_settime\n\t"
    "addq.l #4,sp\n\t"
    "rte"
);

/* Running under EmuTOS? Check OS header.
   Must be run in supervisor mode. */
static ULONG is_emutos(void)
{
    OSHEADER *os_header = *((OSHEADER **)_sysbase);
    return ((ULONG)(os_header->p_rsv2)) == 0x45544F53ul; /* "ETOS" */
}


/* In libcmini */
extern unsigned long _PgmSize;

/* In the inline assembler code */
extern unsigned long xbios_old_trap;
extern void xbios_trap(void);

int main(void)
{
    ULONG t;

    if (Supexec(is_emutos) != 0)
    {
        Cconws("This program is not required under EmuTOS.\r\n");
        return 1;
    }

    if (Supexec(rtc_detect) == 0)
    {
        t = rtc_gettime(TRUE);
        /* Set GEMDOS time. */
        Tsetdate(t >> 16);
        Tsettime(t & 0xFFFFu);

        /* Install XBIOS handler */
        xbios_old_trap = (unsigned long)Setexc(0xB8/4, xbios_trap);

        Cconws("Storm ST RTC driver installed\r\n");
        Ptermres(_PgmSize, 0);
        return 0;
    }
    else
    {
        Cconws("No RTC detected!\r\n");
        return 1;
    }

}